# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from threading import BoundedSemaphore

import structlog

from orchestrator.db import ProductTable, SubscriptionTable
from orchestrator.schedules.scheduling import scheduler
from orchestrator.services.processes import start_process
from orchestrator.services.subscriptions import TARGET_DEFAULT_USABLE_MAP, WF_USABLE_MAP
from orchestrator.targets import Target

logger = structlog.get_logger(__name__)


task_semaphore = BoundedSemaphore(value=2)


@scheduler(name="Subscriptions Validator", time_unit="day", at="00:10")
def validate_subscriptions() -> None:
    subscriptions = SubscriptionTable.query.join(ProductTable).filter(SubscriptionTable.insync.is_(True)).all()
    for subscription in subscriptions:
        validation_workflow = None

        for workflow in subscription.product.workflows:
            if workflow.target == Target.SYSTEM:
                validation_workflow = workflow.name

        if validation_workflow:
            default = TARGET_DEFAULT_USABLE_MAP[Target.SYSTEM]
            usable_when = WF_USABLE_MAP.get(validation_workflow, default)

            if subscription.status in usable_when:
                task_semaphore.acquire()
                json = [{"subscription_id": str(subscription.subscription_id)}]
                _, handle = start_process(validation_workflow, json)
                handle.add_done_callback(lambda _: task_semaphore.release())
        else:
            logger.warning(
                "SubscriptionTable has no validation workflow",
                subscription=subscription,
                product=subscription.product.name,
            )



import pytest

from orchestrator.db import ResourceTypeTable


@pytest.fixture
def resource_type_list():
    return ResourceTypeTable(resource_type="list_field", description="")


@pytest.fixture
def resource_type_int():
    return ResourceTypeTable(resource_type="int_field", description="")


@pytest.fixture
def resource_type_int_2():
    return ResourceTypeTable(resource_type="int_field_2", description="")


@pytest.fixture
def resource_type_str():
    return ResourceTypeTable(resource_type="str_field", description="")



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import requests
import structlog
from requests.exceptions import RequestException
from sqlalchemy.exc import SQLAlchemyError

from orchestrator.db import EngineSettingsTable, ProcessTable, db
from orchestrator.schemas.engine_settings import EngineSettingsSchema
from orchestrator.services.processes import SYSTEM_USER, resume_process
from orchestrator.settings import app_settings

logger = structlog.get_logger(__name__)


def post_update_to_slack(engine_status: EngineSettingsSchema, user: str) -> None:
    """
    Post engine settings update to slack.

    Args:
        engine_status: EngineStatus

    Returns:
        None

    """
    try:
        if engine_status.global_lock is True:
            action = f"stopped the `{app_settings.ENVIRONMENT}` workflow engine. The orchestrator will pause all running processes."
        else:
            action = f"started the `{app_settings.ENVIRONMENT}` workflow engine. The orchestrator will pick up all pending processes."

        message = {"text": f"User `{user}` {action}"}
        requests.post(app_settings.SLACK_ENGINE_SETTINGS_HOOK_URL, json=message)

    # Catch all Request exceptions and log. Then pass
    except RequestException:
        logger.exception("Post to slack failed.")
        pass


def marshall_processes(engine_settings: EngineSettingsTable, new_global_lock: bool) -> Optional[EngineSettingsTable]:
    """
    Manage processes depending on the engine status.

    This function only has to act when in the transitioning fases, i.e Pausing and Starting

    Args:
        engine_settings: Engine status containing the lock and status fields
        new_global_lock: The state to which needs to be transitioned

    Returns:
        Engine status or none

    """
    try:
        # This is the first process/container that will pick up the "running" queue
        if engine_settings.global_lock and not new_global_lock:
            # Update the global lock to unlocked, to make sure no one else picks up the queue
            engine_settings.global_lock = new_global_lock
            db.session.commit()

            # Resume all the running processes
            running_processes = ProcessTable.query.filter(ProcessTable.last_status == "running").all()
            for process in running_processes:
                resume_process(process, user=SYSTEM_USER)

        elif not engine_settings.global_lock and new_global_lock:
            # Lock the engine
            logger.info("Locking the orchestrator engine, Processes will run until the next step")
            engine_settings.global_lock = new_global_lock
            db.session.commit()
        else:
            logger.info(
                "Engine is already locked or unlocked, global lock is unchanged",
                global_lock=engine_settings.global_lock,
                new_status=new_global_lock,
            )

        return engine_settings

    except SQLAlchemyError:
        logger.exception("Encountered a database error, aborting and stopping. Health check will crash the app")
        return None
    except ValueError:
        logger.exception("Encountered an anomaly, locking the engine; manual intervention necessary to fix")
        engine_settings.global_lock = True
        db.session.commit()
        return None



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from copy import deepcopy
from typing import Any, Dict, Generator, List, Optional, TypedDict, Union, cast

import structlog
from pydantic.error_wrappers import ValidationError, display_errors
from pydantic.fields import Field, ModelField, Undefined
from pydantic.main import BaseModel, Extra  # type: ignore

from orchestrator.types import JSON, InputForm, State, StateInputFormGenerator
from orchestrator.utils.json import json_dumps, json_loads

logger = structlog.get_logger(__name__)


__all__ = ("generate_form", "post_process", "FormNotCompleteError", "FormValidationError", "FormException")


class FormException(Exception):
    pass


class FormNotCompleteError(FormException):
    form: InputForm

    def __init__(self, form: JSON):
        super().__init__(form)
        self.form = form


class PydanticErrorDict(TypedDict):
    loc: List[Union[str, int]]
    type: str
    msg: str
    ctx: Dict[str, Any]


class FormValidationError(FormException):
    validator_name: str
    errors: List[PydanticErrorDict]

    def __init__(self, validator_name: str, errors: List[Dict[str, Any]]):
        super().__init__(validator_name, errors)
        self.validator_name = validator_name
        self.errors = cast(List[PydanticErrorDict], errors)

    def __str__(self) -> str:
        no_errors = len(self.errors)
        return (
            f'{no_errors} validation error{"" if no_errors == 1 else "s"} for {self.validator_name}\n'
            f"{display_errors(cast(List[Dict[str, Any]], self.errors))}"  # type: ignore
        )


def generate_form(
    form_generator: Optional[StateInputFormGenerator], state: State, user_inputs: List[State]
) -> Optional[InputForm]:
    """Generate form using form generator as defined by a workflow."""
    try:
        # Generate form is basically post_process
        post_process(form_generator, state, user_inputs)
    except FormNotCompleteError as e:
        # Form is not finished and raises the next form, this is expected
        return e.form

    # Form is finished and thus there is no new form
    return None


def post_process(form_generator: Optional[StateInputFormGenerator], state: State, user_inputs: List[State]) -> State:
    """Post process user_input based on form definition from workflow."""

    # there is no form_generator so we return no validated data
    if not form_generator:
        return {}

    current_state = deepcopy(state)

    logger.debug("Post process form", state=state, user_inputs=user_inputs)

    # Generate generator
    generator = form_generator(current_state)

    try:
        # Generate first form (we need to send None here, since the arguments are already given when we generated the generator)
        generated_form: InputForm = generator.send(None)

        # Loop through user inputs and for each input validate and update current state and validation results
        for user_input in user_inputs:
            # Validate
            try:
                form_validated_data = generated_form(**user_input)
            except ValidationError as e:
                raise FormValidationError(e.model.__name__, e.errors()) from e  # type: ignore

            # Update state with validated_data
            current_state.update(form_validated_data.dict())

            # Make next form
            generated_form = generator.send(form_validated_data)

        else:
            # Form is not completely filled raise next form
            raise FormNotCompleteError(generated_form.schema())
    except StopIteration as e:
        # Form is completely filled so we can return the last of the data and
        return e.value


class DisplayOnlyFieldType:
    @classmethod
    def __get_validators__(cls) -> Generator:
        yield cls.nothing

    def nothing(cls, v: Any, field: ModelField) -> Any:
        return field.default


class FormPage(BaseModel):
    class Config:
        json_loads = json_loads
        json_dumps = json_dumps
        title = "unknown"
        extra = Extra.forbid
        validate_all = True

    def __init_subclass__(cls, /, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)  # type:ignore

        # The default and requiredness of a field is not a property of a field
        # In the case of DisplayOnlyFieldTypes, we do kind of want that.
        # Using this method we set the right properties after the form is created
        for field in cls.__fields__.values():
            try:
                if issubclass(field.type_, DisplayOnlyFieldType):
                    field.required = False
                    field.allow_none = True
            except TypeError:
                pass


def ReadOnlyField(
    default: Any = Undefined,
    *,
    const: Optional[bool] = None,
    **extra: Any,
) -> Any:
    return Field(default, const=True, uniforms={"disabled": True, "value": default}, **extra)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, Dict, Iterator, Optional

from pydantic import BaseModel, ConstrainedInt, conint, root_validator

from orchestrator.utils.vlans import VlanRanges


class BFD(BaseModel):
    class Config:
        schema_extra = {"format": "optGroup"}

    # order matters, this should be first
    enabled: bool
    minimum_interval: Optional[conint(ge=1, le=255000)] = 900  # type: ignore
    multiplier: Optional[conint(ge=1, le=255)] = 3  # type: ignore

    @root_validator()
    def check_optional_fields(cls, values: Dict) -> Dict:  # noqa: B902
        if not values.get("enabled"):
            values.pop("minimum_interval", None)
            values.pop("multiplier", None)

        return values


class MTU(ConstrainedInt):
    ge = 1500
    le = 9000

    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
        super().__modify_schema__(field_schema)
        field_schema["multipleOf"] = 7500


class VlanRangesValidator(VlanRanges):
    @classmethod
    def __modify_schema__(cls, field_schema: Dict) -> None:
        field_schema.update(
            pattern="^([1-4][0-9]{0,3}(-[1-4][0-9]{0,3})?,?)+$",
            examples=["345", "20-23,45,50-100"],
            type="string",
            format="vlan",
        )

    @classmethod
    def __get_validators__(cls) -> Iterator:
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield VlanRanges



from test.unit_tests.domain.products.product_blocks.product_block_list_nested import (  # noqa: F401
    test_product_block_list_nested,
    test_product_block_list_nested_db_in_use_by_block,
)
from test.unit_tests.domain.products.product_blocks.product_block_one import (  # noqa: F401
    test_product_block_one,
    test_product_block_one_db,
)
from test.unit_tests.domain.products.product_blocks.product_block_one_nested import (  # noqa: F401
    test_product_block_one_nested,
    test_product_block_one_nested_db_in_use_by_block,
)
from test.unit_tests.domain.products.product_blocks.product_block_with_list_union import (  # noqa: F401
    test_product_block_with_list_union,
    test_product_block_with_list_union_db,
)
from test.unit_tests.domain.products.product_blocks.product_block_with_union import (  # noqa: F401
    test_product_block_with_union,
    test_product_block_with_union_db,
)
from test.unit_tests.domain.products.product_blocks.product_sub_block_one import (  # noqa: F401
    test_product_sub_block_one,
    test_product_sub_block_one_db,
)
from test.unit_tests.domain.products.product_blocks.product_sub_block_two import (  # noqa: F401
    test_product_sub_block_two,
    test_product_sub_block_two_db,
)
from test.unit_tests.domain.products.product_types.product_type_list_nested import (  # noqa: F401
    test_product_list_nested,
    test_product_model_list_nested,
    test_product_type_list_nested,
)
from test.unit_tests.domain.products.product_types.product_type_list_union import (  # noqa: F401
    test_product_list_union,
    test_product_type_list_union,
)
from test.unit_tests.domain.products.product_types.product_type_list_union_overlap import (  # noqa: F401
    test_product_list_union_overlap,
    test_product_type_list_union_overlap,
)
from test.unit_tests.domain.products.product_types.product_type_one import (  # noqa: F401
    test_product_model,
    test_product_one,
    test_product_type_one,
)
from test.unit_tests.domain.products.product_types.product_type_one_nested import (  # noqa: F401
    test_product_model_nested,
    test_product_one_nested,
    test_product_type_one_nested,
)
from test.unit_tests.domain.products.product_types.product_type_sub_list_union import (  # noqa: F401
    test_product_sub_list_union,
    test_product_type_sub_list_union,
)
from test.unit_tests.domain.products.product_types.product_type_sub_one import (  # noqa: F401
    sub_one_subscription_1,
    test_product_sub_one,
    test_product_type_sub_one,
)
from test.unit_tests.domain.products.product_types.product_type_sub_two import (  # noqa: F401
    sub_two_subscription_1,
    test_product_sub_two,
    test_product_type_sub_two,
)
from test.unit_tests.domain.products.product_types.product_type_sub_union import (  # noqa: F401
    test_union_sub_product,
    test_union_type_sub_product,
)
from test.unit_tests.domain.products.product_types.product_type_union import (  # noqa: F401
    test_union_product,
    test_union_type_product,
)
from test.unit_tests.domain.products.resource_types import (  # noqa: F401
    resource_type_int,
    resource_type_int_2,
    resource_type_list,
    resource_type_str,
)
